/*
 *
 * Copyright (c) 2003-2004 Erez Zadok
 * Copyright (c) 2003-2004 Swapnil V. Patil
 * Copyright (c) 2003-2004 Anand Kashyap
 * Copyright (c) 2003-2004 Gopalan Sivathanu
 * Copyright (c) 2003-2004 Stony Brook University
 * Copyright (c) 2003-2004 The Research Foundation of State University of New York
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package.
 *
 * This Copyright notice must be kept intact and distributed with all sources.
 */

/*
 *  $Id: file.c,v 1.28 2004/08/19 23:40:48 gopalan Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#ifdef FISTGEN
# include "fist_i3fs.h"
#endif /* FISTGEN */
#include "fist.h"
#include "i3fs.h"
#include "i3fs_utils.h"


/* File Operations */

STATIC loff_t
i3fs_llseek(file_t *file, loff_t offset, int origin)
{
    loff_t err;
    file_t *hidden_file = NULL;

    print_entry_location();
    if (ftopd(file) != NULL) {
	hidden_file = ftohf(file);
    }

    fist_dprint(6, "i3fs_llseek: file=%p, offset=0x%x, origin=%d\n",
		file, offset, origin);

    ASSERT(hidden_file != NULL);

    /* always set hidden position to this one */
    hidden_file->f_pos = file->f_pos;
    if (file->f_reada) { /* update readahead information if needed */
	hidden_file->f_reada = file->f_reada;
	hidden_file->f_ramax = file->f_ramax;
	hidden_file->f_raend = file->f_raend;
	hidden_file->f_ralen = file->f_ralen;
	hidden_file->f_rawin = file->f_rawin;
    }

    if (hidden_file->f_op && hidden_file->f_op->llseek)
	err = hidden_file->f_op->llseek(hidden_file, offset, origin);
    else
	err = generic_file_llseek(hidden_file, offset, origin);

    if (err < 0)
	goto out;

    if (err != file->f_pos) {
	file->f_pos = err;
	// ION maybe this?
	// 	file->f_pos = hidden_file->f_pos;
	file->f_reada = 0;
	file->f_version = ++event;
    }
out:
    print_exit_status(err);
    return err;
}


/*
 * generic_file_read updates the atime of upper layer inode.
 * But, it doesn't give us a chance to update the atime of
 * the lower layer inode. This function is a wrapper to
 * generic_file_read. it Updates the atime of the lower level
 * inode if generic_file_read returns without any errors
 * This is to be used only for file reads and only if
 * FIST_FILTER_DATA is defined. The function to be
 * used for directory reads is i3fs_read
 */
STATIC ssize_t
i3fs_read_update_atime(file_t *file, char* buf, size_t count, loff_t *ppos)
{
    int err = 0;

    print_entry_location();

    err = generic_file_read(file, buf, count, ppos);

    if (err >= 0)
	UPDATE_ATIME(itohi(file->f_dentry->d_inode));

    print_exit_status(err);
    return err;
}

STATIC ssize_t
i3fs_read(file_t *file, char *buf, size_t count, loff_t *ppos)
{
    int err = -EINVAL;
    file_t *hidden_file = NULL;
    loff_t pos = *ppos;

    print_entry_location();
    if (ftopd(file) != NULL)
	hidden_file = ftohf(file);
    ASSERT(hidden_file != NULL);

    if (!hidden_file->f_op || !hidden_file->f_op->read)
	goto out;

    err = hidden_file->f_op->read(hidden_file, buf, count, &pos);

    if (err >= 0) {
	/* atime should also be updated for reads of size zero or more */
	fist_copy_attr_atime(file->f_dentry->d_inode,
			     hidden_file->f_dentry->d_inode);
    }

    // MAJOR HACK
    /*
     * because pread() does not have any way to tell us that it is
     * our caller, then we don't know for sure if we have to update
     * the file positions.  This hack relies on read() having passed us
     * the "real" pointer of its struct file's f_pos field.
     */
    if (ppos == &file->f_pos)
	hidden_file->f_pos = *ppos = pos;
    if (hidden_file->f_reada) { /* update readahead information if needed */
	file->f_reada = hidden_file->f_reada;
	file->f_ramax = hidden_file->f_ramax;
	file->f_raend = hidden_file->f_raend;
	file->f_ralen = hidden_file->f_ralen;
	file->f_rawin = hidden_file->f_rawin;
    }

out:
    print_exit_status(err);
    return err;
}


/* this i3fs_write() does not modify data pages! */
STATIC ssize_t
i3fs_write(file_t *file, const char *buf, size_t count, loff_t *ppos)
{
    int err = -EINVAL;
    file_t *hidden_file = NULL;
    inode_t *inode;
    inode_t *hidden_inode;
    loff_t pos = *ppos;

    print_entry_location();
    if (ftopd(file) != NULL)
	hidden_file = ftohf(file);
    ASSERT(hidden_file != NULL);

    inode = file->f_dentry->d_inode;
    hidden_inode = itohi(inode);

    if (!hidden_file->f_op || !hidden_file->f_op->write)
	goto out;

    /* adjust for append -- seek to the end of the file */
    if (file->f_flags & O_APPEND)
	pos = inode->i_size;

    err = hidden_file->f_op->write(hidden_file, buf, count, &pos);

    /*
     * copy ctime and mtime from lower layer attributes
     * atime is unchanged for both layers
     */
    if (err >= 0)
	fist_copy_attr_times(inode, hidden_inode);

    /*
     * XXX: MAJOR HACK
     *
     * because pwrite() does not have any way to tell us that it is
     * our caller, then we don't know for sure if we have to update
     * the file positions.  This hack relies on write() having passed us
     * the "real" pointer of its struct file's f_pos field.
     */
    if (ppos == &file->f_pos)
	hidden_file->f_pos = *ppos = pos;

    /* update this inode's size */
    if (pos > inode->i_size)
	inode->i_size = pos;

out:
    print_exit_status(err);
    return err;
}


#if defined(FIST_FILTER_NAME) || defined(FIST_FILTER_SCA)
struct i3fs_getdents_callback {
    void *dirent;
    dentry_t *dentry;
    filldir_t filldir;
    int err;
    int filldir_called;
    int entries_written;
};

/* copied from generic filldir in fs/readir.c */
STATIC int
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,9)
i3fs_filldir(void *dirent, const char *name, int namlen, loff_t offset, ino_t ino, unsigned int d_type)
# else /* kernel older than 2.4.9 */
i3fs_filldir(void *dirent, const char *name, int namlen, off_t offset, ino_t ino, unsigned int d_type)
# endif /* kernel older than 2.4.9 */
{
    struct i3fs_getdents_callback *buf = (struct i3fs_getdents_callback *) dirent;
    int err = 0;
    char *decoded_name;
    int decoded_length;

    print_entry_location();

    buf->filldir_called++;


    decoded_length = i3fs_decode_filename(name, namlen, &decoded_name, SKIP_DOTS,
					    buf->dentry->d_inode, buf->dentry->d_sb);
    if (decoded_length < 0) {
	return 0;			/* no error, just skip the entry */
    }

	/* Hiding the Database File */
	if(!strncmp(decoded_name, DATABASE_FILE, decoded_length)) return err;

    err = buf->filldir(buf->dirent, decoded_name, decoded_length, offset, ino, d_type);
    KFREE(decoded_name);
    if (err >= 0) {
    	buf->entries_written++;
    }

out:
    print_exit_status(err);
    return err;
}
#endif /* FIST_FILTER_NAME || FIST_FILTER_SCA */


STATIC int
i3fs_readdir(file_t *file, void *dirent, filldir_t filldir)
{
    int err = -ENOTDIR;
    file_t *hidden_file = NULL;
    inode_t *inode;
#if defined(FIST_FILTER_NAME) || defined(FIST_FILTER_SCA)
    struct i3fs_getdents_callback buf;
#endif /* FIST_FILTER_NAME || FIST_FILTER_SCA */

    print_entry_location();
    if (ftopd(file) != NULL)
	hidden_file = ftohf(file);
    ASSERT(hidden_file != NULL);

    inode = file->f_dentry->d_inode;

    fist_checkinode(inode, "i3fs_readdir");
    hidden_file->f_pos = file->f_pos;

#if defined(FIST_FILTER_NAME) || defined(FIST_FILTER_SCA)
    /* prepare for callback */
    buf.dirent = dirent;
    buf.dentry = file->f_dentry;
    buf.filldir = filldir;
retry:
    buf.filldir_called = 0;
    buf.entries_written = 0;
    buf.err = 0;
    err = vfs_readdir(hidden_file, i3fs_filldir, (void *) &buf);
    if (buf.err) {
	err = buf.err;
    }
    if (buf.filldir_called && !buf.entries_written) {
	goto retry;
    }
#else /* not FIST_FILTER_NAME || FIST_FILTER_SCA */
    err = vfs_readdir(hidden_file, filldir, dirent);
#endif /* not FIST_FILTER_NAME || FIST_FILTER_SCA */

    file->f_pos = hidden_file->f_pos;
    if (err > 0)
	fist_copy_attr_atime(inode, hidden_file->f_dentry->d_inode);

    fist_checkinode(inode, "post i3fs_readdir");
    print_exit_status(err);
    return err;
}


STATIC unsigned int
i3fs_poll(file_t *file, poll_table *wait)
{
    unsigned int mask = DEFAULT_POLLMASK;
    file_t *hidden_file = NULL;

    print_entry_location();
    if (ftopd(file) != NULL)
	hidden_file = ftohf(file);
    ASSERT(hidden_file != NULL);

    if (!hidden_file->f_op || !hidden_file->f_op->poll)
	goto out;

    mask = hidden_file->f_op->poll(hidden_file, wait);

out:
    print_exit_status(mask);
    return mask;
}


/* Authenticate the ioctl */

int i3fs_authenticate(struct super_block *sb, struct policy *arg)	{
	if (strcmp(arg->passphrase, stopd(sb)->passphrase) == 0)
		return 1;
	else
		return 0;
}


STATIC int
i3fs_ioctl(inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg)
{
    int err = 0;		/* don't fail by default */
    file_t *hidden_file = NULL;
    vfs_t *this_vfs;
    vnode_t *this_vnode;
    int val = 0;
    long checksum;
#ifdef FIST_COUNT_WRITES
    extern unsigned long count_writes, count_writes_middle;
#endif /* FIST_COUNT_WRITES */

    print_entry_location();

    this_vfs = inode->i_sb;
    this_vnode = inode;

    /* check if asked for local commands */
    switch (cmd) {
    case FIST_IOCTL_GET_DEBUG_VALUE:
	val = fist_get_debug_value();
	printk("IOCTL GET: send arg %d\n", val);
	err = put_user(val, (int *) arg);
#ifdef FIST_COUNT_WRITES
	printk("COUNT_WRITES:%lu:%lu\n", count_writes, count_writes_middle);
# endif /* FIST_COUNT_WRITES */
	break;

    case FIST_IOCTL_SET_DEBUG_VALUE:
	err = get_user(val, (int *) arg);
	if (err)
	    break;
	fist_dprint(6, "IOCTL SET: got arg %d\n", val);
	if (val < 0 || val > 20) {
	    err = -EINVAL;
	    break;
	}
	fist_set_debug_value(val);
	break;

/* policy specific ioctls begin here.  For all these ioctls, authentication
 * has to be done first */

    case ADD_POLICY:
	if (!i3fs_authenticate(inode->i_sb, (struct policy*)arg))	{
		err = -EACCES;
		goto out;
	}

	put_policy_in_database(inode->i_sb, (struct policy*)arg);
	break;

    case REMOVE_POLICY:
        if (!i3fs_authenticate(inode->i_sb, (struct policy*)arg))      {
                err = -EACCES;
                goto out;
        }

	remove_policy(((struct policy*)arg)->file);	
	break;


    case ALLOW_UPDATES:
	if (!i3fs_authenticate(inode->i_sb, (struct policy*)arg))      {
                err = -EACCES;
		printk(KERN_WARNING "\nAccess denied for ioctl");
                goto out;
        }
	printk("\nAllowing updates");

	stopd(inode->i_sb)->allow_updates = 1;

	break;

    case DISALLOW_UPDATES:
	if (!i3fs_authenticate(inode->i_sb, (struct policy*)arg))      {
                err = -EACCES;
                goto out;
        }
	printk("\nDisAllowing updates");

	stopd(inode->i_sb)->allow_updates = 0;
	break;

    default:
        if (ftopd(file) != NULL) {
	    hidden_file = ftohf(file);
	}
	/* pass operation to hidden filesystem, and return status */
	if (hidden_file && hidden_file->f_op && hidden_file->f_op->ioctl)
	    err = hidden_file->f_op->ioctl(itohi(inode), hidden_file, cmd, arg);
	else
	    err = -ENOTTY;	/* this is an unknown ioctl */
    } /* end of outer switch statement */


out:
    print_exit_status(err);
    return err;
}


/* to verify checksum match for the entire file */

int
verify_filedata_integrity(dentry_t *dentry, super_block_t *sb)	{
	char *computed_checksum, *stored_checksum;
	int ret=0;
	file_t *fp;
	char *path, *tmp;
	
	computed_checksum = KMALLOC(CHECKSUM_SIZE, GFP_KERNEL);
	stored_checksum = KMALLOC(CHECKSUM_SIZE, GFP_KERNEL);

	memset(computed_checksum, 0, CHECKSUM_SIZE);
	memset(stored_checksum, 0, CHECKSUM_SIZE);

	if (!computed_checksum || !stored_checksum)	{
		ret = -ENOMEM;
		goto out;
	}


	dget(dentry);

	path = (char*)KMALLOC(PAGE_SIZE, GFP_KERNEL);

	/* we need the path name here for opening the file */

	tmp = d_path(dtohd(dentry), stopd(sb)->hidden_mnt, path, PAGE_SIZE);
	fp = filp_open(tmp, O_RDONLY, 0);

	compute_filedata_checksum(fp, sb, computed_checksum);

	filp_close(fp, NULL);
        KFREE(path);

	get_stored_data_checksum(stopd(sb)->datadb, itohi(dentry->d_inode)->i_ino, 0, stored_checksum);

	if (memcmp(computed_checksum, stored_checksum, CHECKSUM_SIZE) != 0)
		ret = -EACCES;
	else	{
		ret = 1;
	//	printk("\nWhole file checksumming verified.");
	}
	
	dput(dentry);

out:
	KFREE(computed_checksum);
	KFREE(stored_checksum);
	return ret;
}
	



STATIC int
i3fs_open(inode_t *inode, file_t *file)
{
    int err = 0, bits=0, ret=0;
    int hidden_mode, hidden_flags;
    file_t *hidden_file = NULL;
    dentry_t *hidden_dentry = NULL;

    print_entry_location();
    ftopd(file) = KMALLOC(sizeof(struct i3fs_file_info), GFP_KERNEL);
    if (!ftopd(file)) {
        err = -ENOMEM;
        goto out;
    }

    hidden_dentry = i3fs_hidden_dentry(file->f_dentry); /* CPW: Moved after print_entry */
    fist_print_dentry("i3fs_open IN hidden_dentry", hidden_dentry);

    dget(hidden_dentry);

    hidden_flags = file->f_flags;
    if ((hidden_flags & O_ACCMODE) == O_WRONLY)
        hidden_flags = (hidden_flags & O_ACCMODE) | O_RDWR;
    if (file->f_flags & O_APPEND) {
        fist_dprint(5, "file is opened in append-only mode\n");
        hidden_flags &= ~O_APPEND; /* turn off O_APPEND flag */
    }

    /*
     * dentry_open will decrement mnt refcnt if err.
     * otherwise fput() will do an mntput() for us upon file close.
     */
    mntget(stopd(inode->i_sb)->hidden_mnt);
    hidden_file = dentry_open(hidden_dentry, stopd(inode->i_sb)->hidden_mnt, hidden_flags);
    if (IS_ERR(hidden_file)) {
        err = PTR_ERR(hidden_file);
        goto out;
    }

    ftohf(file) = hidden_file;  /* link two files */

// integrity checking if the checksum policy is to checksum the entire file.


    if ( itopd(inode)->policy != 0 && !S_ISDIR(inode->i_mode) && itopd(inode)->check_required)	{
	if ((itopd(inode)->policy & DATA) && !(itopd(inode)->policy & PER_PAGE))
		/* Check if there is a custom frequency of checks */
		if (frequency_pass(inode))	{
			if (stopd(inode->i_sb)->cache_results)	{

				/* Check cache */

				if (itopd(inode)->whole_file_result == 1)
					goto out;

				if (itopd(inode)->whole_file_result == 2)	{
					itopd(inode)->whole_file_result = 2;
					report_data_mismatch(file->f_dentry, current);
	                                dput(hidden_dentry);
	                                KFREE(ftopd(file));
	                                fput(hidden_file);
	                                return -EPERM;
				}
			}
		
	
			ret = verify_filedata_integrity(file->f_dentry, file->f_dentry->d_inode->i_sb);

			if (ret < 0)	{
				/* Cache the result */
				
				itopd(inode)->whole_file_result = 2;
				report_data_mismatch(file->f_dentry, current);
				if (itopd(inode)->policy & BLOCK)	{
					dput(hidden_dentry);
					KFREE(ftopd(file));
					fput(hidden_file);
					return ret;
				}
			} else	{
				/* Cache the result */
				
				itopd(inode)->whole_file_result = 1;
			}

		}
    }

out:
    if (err < 0 && ftopd(file)) {
        KFREE(ftopd(file));
    }
    fist_print_dentry("i3fs_open OUT hidden_dentry", hidden_dentry);
    print_exit_status(err);
    return err;
}


STATIC int
i3fs_flush(file_t *file)
{
    int err = 0, bits=0;		/* assume ok (see open.c:close_fp) */
    file_t *hidden_file = NULL;

    print_entry_location();

    if (ftopd(file) != NULL)
	hidden_file = ftohf(file); /* CPW: Moved after print_entry_location */
    ASSERT(hidden_file != NULL);

    if (!hidden_file->f_op || !hidden_file->f_op->flush)
	goto out;

    err = hidden_file->f_op->flush(hidden_file);

out:
    print_exit_status(err);
    return err;
}


STATIC int
i3fs_release(inode_t *inode, file_t *file)
{
    int err = 0, bits=0;
    file_t *hidden_file = NULL, *fp = NULL;
    dentry_t *hidden_dentry;
    char *path, *tmp;

    print_entry_location();


// Update the whole file checksum if updates are allowed

    if (ftopd(file) != NULL)    {
        if ( (bits = itopd(file->f_dentry->d_inode)->policy) != 0)      {
                if (stopd(inode->i_sb)->allow_updates && (bits & DATA) && !(bits & PER_PAGE))       {
			path = (char*)KMALLOC(PAGE_SIZE, GFP_KERNEL);
		        tmp = d_path(dtohd(file->f_dentry), stopd(inode->i_sb)->hidden_mnt, path, PAGE_SIZE);
			fp = filp_open(tmp, O_RDONLY, 0);
			if (fp && !IS_ERR(fp))	{
	                        checksum_filedata_to_db(file->f_dentry, fp);
				filp_close(fp, NULL);
			}
		        KFREE(path);
                }
        }
    }



    if (ftopd(file) != NULL)
	hidden_file = ftohf(file);
    KFREE(ftopd(file));

    ASSERT(hidden_file != NULL);

    fist_print_dentry("i3fs_release IN hidden_dentry", hidden_file->f_dentry);
    fist_checkinode(inode, "i3fs_release");
    fist_dprint(6, "i3fs_release IN, file->f_count=%d\n", file->f_count);
    /*
     * will decrement file refcount, and if 0, destroy the file,
     * which will call the lower file system's file release function.
     */
    hidden_dentry = hidden_file->f_dentry;
    fput(hidden_file);

    fist_dprint(6, "i3fs_release done\n");
    fist_checkinode(inode, "post i3fs_release");

    fist_print_dentry("i3fs_release OUT hidden_dentry", hidden_dentry);
    print_exit_status(err);
    return err;
}


STATIC int
i3fs_fsync(file_t *file, dentry_t *dentry, int datasync)
{
    int err = -EINVAL;
    file_t *hidden_file = NULL;
    dentry_t *hidden_dentry;

    print_entry_location();
    if (ftopd(file) != NULL)
	hidden_file = ftohf(file);
    ASSERT(hidden_file != NULL);
    hidden_dentry = i3fs_hidden_dentry(dentry);

    if (hidden_file->f_op && hidden_file->f_op->fsync) {
	down(&hidden_dentry->d_inode->i_sem);
	err = hidden_file->f_op->fsync(hidden_file, hidden_dentry, datasync);
	up(&hidden_dentry->d_inode->i_sem);
    }

    print_exit_status(err);
    return err;
}


STATIC int
i3fs_fasync(int fd, file_t *file, int flag)
{
    int err = 0;
    file_t *hidden_file = NULL;

    print_entry_location();
    if (ftopd(file) != NULL)
	hidden_file = ftohf(file);
    ASSERT(hidden_file != NULL);

    if (hidden_file->f_op && hidden_file->f_op->fasync)
	err = hidden_file->f_op->fasync(fd, hidden_file, flag);

    print_exit_status(err);
    return err;
}


STATIC int
i3fs_lock(file_t *file, int cmd, struct file_lock *fl)
{
    int err = 0;
    file_t *hidden_file = NULL;
    struct file_lock *tmpfl = NULL;

    print_entry_location();
    if (ftopd(file) != NULL)
	hidden_file = ftohf(file);
    ASSERT(hidden_file != NULL);

    if (hidden_file->f_op->lock) {
	/*
	 * XXX: fl->fl_file is a copy of our own struct file pointer!  It's
	 * bad to pass that to the lower f/s: NFS in particular causes an
	 * oops deep inside the NLM code.  So the hack here is to
	 * temporarily fix the fl->fl_file field to the hidden_file, and fix
	 * it back after returning form ->lock().  This hack may not be
	 * race-free or SMP-safe.  Maybe we need to create a private struct
	 * file_lock, carefully initialize it, and then pass that down?  We
	 * may also need to replicate much of fs/locks.c:fcntl_getlk().
	 */
	fl->fl_file = hidden_file;
	err = hidden_file->f_op->lock(hidden_file, F_GETLK, fl);
	fl->fl_file = file;
	if (err < 0)
	    goto out;
	else if (err == LOCK_USE_CLNT)
	    /* Bypass for NFS with no locking - 2.0.36 compat */
	    tmpfl = posix_test_lock(hidden_file, fl);
    } else {
	tmpfl = posix_test_lock(hidden_file, fl);
    }
    /*
     * XXX: based on fs/locks.c: fcntl_getlk(), I'm not sure this is correct.
     */
    if (tmpfl)
	err = 0;	       /* all OK from above posix_test_lock calls */
    else
	err = -EINVAL;

out:
    print_exit_status(err);
    return err;
}


struct file_operations i3fs_dir_fops =
{
    llseek:	i3fs_llseek,
    read:	i3fs_read,
    write:	i3fs_write,
    readdir:	i3fs_readdir,
    poll:	i3fs_poll,
    ioctl:	i3fs_ioctl,
    mmap:	generic_file_mmap,
    open:	i3fs_open,
    flush:	i3fs_flush,
    release:	i3fs_release,
    fsync:	i3fs_fsync,
    fasync:	i3fs_fasync,
    lock:	i3fs_lock,
    /* not needed: readv */
    /* not needed: writev */
    /* not implemented: sendpage */
    /* not implemented: get_unmapped_area */
};

struct file_operations i3fs_main_fops =
{
    llseek:	i3fs_llseek,
    read:	i3fs_read_update_atime,
    write:	generic_file_write,
    readdir:	i3fs_readdir,
    poll:	i3fs_poll,
    ioctl:	i3fs_ioctl,
    mmap:	generic_file_mmap,
    open:	i3fs_open,
    flush:	i3fs_flush,
    release:	i3fs_release,
    fsync:	i3fs_fsync,
    fasync:	i3fs_fasync,
    lock:	i3fs_lock,
    /* not needed: readv */
    /* not needed: writev */
    /* not implemented: sendpage */
    /* not implemented: get_unmapped_area */
};

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
