/*
 *
 * Copyright (c) 2003-2004 Erez Zadok
 * Copyright (c) 2003-2004 Swapnil V. Patil
 * Copyright (c) 2003-2004 Anand Kashyap
 * Copyright (c) 2003-2004 Gopalan Sivathanu
 * Copyright (c) 2003-2004 Stony Brook University
 * Copyright (c) 2003-2004 The Research Foundation of State University of New York
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package.
 *
 * This Copyright notice must be kept intact and distributed with all sources.
 */

/*
 *  $Id: main.c,v 1.23 2004/08/19 23:40:48 gopalan Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#ifdef FISTGEN
# include "fist_i3fs.h"
#endif /* FISTGEN */
#include "fist.h"
#include "i3fs.h"
#include "i3fs_utils.h"
#include <linux/module.h>

/* This definition must only appear after we include <linux/module.h> */
#ifndef MODULE_LICENSE
# define MODULE_LICENSE(bison)
#endif /* not MODULE_LICENSE */


/* sb we pass is i3fs's super_block */
int
i3fs_interpose(dentry_t *hidden_dentry, dentry_t *dentry, super_block_t *sb, int flag)
{
    inode_t *hidden_inode;
    int err = 0;
    inode_t *inode;

    print_entry_location();

    hidden_inode = hidden_dentry->d_inode; /* CPW: moved after print_entry_location */

    ASSERT(hidden_inode != NULL);
    ASSERT(dentry->d_inode == NULL);

    /*
     * We allocate our new inode below, by calling iget.
     * iget will call our read_inode which will initialize some
     * of the new inode's fields
     */
    /* XXX: fist file systems reserve first 10 inodes */
    inode = iget(sb, i3fs_iunique(sb, 10));

    if (!inode) {
	err = -EACCES;		/* should be impossible??? */
	goto out;
    }

    /*
     * interpose the inode if not already interposed
     *   this is possible if the inode is being reused
     * XXX: what happens if we get_empty_inode() but there's another already?
     * for now, ASSERT() that this can't happen; fix later.
     */
    if (itohi(inode) == NULL)
	itohi(inode) = igrab(hidden_inode);

    /* Use different set of inode ops for symlinks & directories*/
    if (S_ISLNK(hidden_inode->i_mode))
	inode->i_op = &i3fs_symlink_iops;
    else if (S_ISDIR(hidden_inode->i_mode))
	inode->i_op = &i3fs_dir_iops;
    /* Use different set of file ops for directories */
    if (S_ISDIR(hidden_inode->i_mode))
	inode->i_fop = &i3fs_dir_fops;

    /* properly initialize special inodes */
    if (S_ISBLK(hidden_inode->i_mode) || S_ISCHR(hidden_inode->i_mode) ||
	S_ISFIFO(hidden_inode->i_mode) || S_ISSOCK(hidden_inode->i_mode)) {
	init_special_inode(inode, hidden_inode->i_mode, hidden_inode->i_rdev);
    }


    /* only (our) lookup wants to do a d_add */
    if (flag)
	d_add(dentry, inode);
    else
	d_instantiate(dentry, inode);

    ASSERT(dtopd(dentry) != NULL);

    /* all well, copy inode attributes */
    fist_copy_attr_all(inode, hidden_inode);

 out:
    print_exit_status(err);
    return err;
}


/* find hidden dentry given this i3fs dentry */
dentry_t *
__i3fs_hidden_dentry(char *file, char *func, int line, dentry_t *dentry)
{
    dentry_t *hidden_dentry;

    ASSERT2(dentry != NULL);
    ASSERT2(dentry->d_op != NULL);
    ASSERT2(dentry->d_op == &i3fs_dops
	    );
    ASSERT2(dentry->d_sb->s_op == &i3fs_sops);
    if (dentry->d_inode)
	ASSERT2(dentry->d_inode->i_op == &i3fs_main_iops ||
		dentry->d_inode->i_op == &i3fs_dir_iops ||
		dentry->d_inode->i_op == &i3fs_symlink_iops);

    hidden_dentry = dtohd(dentry);
    ASSERT2(hidden_dentry != NULL);
    return hidden_dentry;
}


//set up the four in-kernel berkeley databases.

void setup_databases(super_block_t *sb, char *prefix)	{
    char mount_path[100], *path_buf;
    int path_len, ret=0;
    DB_ENV *database = NULL;
    DB *policydb, *datadb, *metadb, *accessdb;

	if ((ret = db_env_create(&database, 0)) != 0)
		printk(KERN_WARNING "i3fs: env creation failed");
	if ((ret = database->set_encrypt(database, "gopalan", DB_ENCRYPT_AES)) != 0)
		printk(KERN_WARNING "i3fs: setting encryption failed");


	/* Create the databases required */
	if (db_create(&(stopd(sb)->policydb), NULL, 0) != 0)    {
		printk(KERN_WARNING "i3fs: Unable to create policy db");
	}

	if (db_create(&(stopd(sb)->datadb), NULL, 0) != 0)      {
		printk(KERN_WARNING "i3fs: Unable to create data db");
	}

	if (db_create(&(stopd(sb)->metadb), NULL, 0) != 0)      {
		printk(KERN_WARNING "i3fs: Unable to create metadata db");
	}

	if (db_create(&(stopd(sb)->accessdb), NULL, 0) != 0)    {
		printk(KERN_WARNING "i3fs: Unable to create access db");
	}


	policydb = stopd(sb)->policydb;
	datadb = stopd(sb)->datadb;
	metadb = stopd(sb)->metadb;
	accessdb = stopd(sb)->accessdb;

	// set optimal cache size for the databases

	policydb->set_cachesize(policydb, 0, (1024*1024), 0);
	metadb->set_cachesize(metadb, 0, (2*1024*1024), 0);
	datadb->set_cachesize(datadb, 0, (1024*1024), 0);
	accessdb->set_cachesize(accessdb, 0, (1024*512), 0);

	// set optimal page size for the databases

	policydb->set_pagesize(policydb, 4096);
	datadb->set_pagesize(datadb, 4096);
	metadb->set_pagesize(metadb, 4096);
	accessdb->set_pagesize(accessdb, 4096);

	 // set encryption for the databases.-- password should be got as a mount option -- yet to be implemented.

	policydb->set_encrypt(policydb, stopd(sb)->passphrase, DB_ENCRYPT_AES);
	datadb->set_encrypt(datadb, stopd(sb)->passphrase, DB_ENCRYPT_AES);
	metadb->set_encrypt(metadb, stopd(sb)->passphrase, DB_ENCRYPT_AES);
	accessdb->set_encrypt(accessdb, stopd(sb)->passphrase, DB_ENCRYPT_AES);

	// if there is a custom db path, then use it as the prefix
	// if not use the mount directory

	if (prefix)
		strcpy(mount_path, prefix);
	else
		strcpy(mount_path, stopd(sb)->mntdir);

	strcat(mount_path, "/");

	path_len = 1;

	strcat(mount_path, "policy.db");
	if (policydb->open(policydb, NULL, mount_path, NULL, DB_BTREE, DB_CREATE , 0) != 0)     {
		printk(KERN_WARNING "i3fs: Unable to open access db");
	}
	mount_path[path_len] = '\0';

	strcat(mount_path, "data.db");
	if (datadb->open(datadb, NULL, mount_path, NULL, DB_BTREE, DB_CREATE , 0) != 0) {
		printk(KERN_WARNING "i3fs: Unable to open access db");
	}
	mount_path[path_len] = '\0';

	strcat(mount_path, "meta.db");
	if (metadb->open(metadb, NULL, mount_path, NULL, DB_BTREE, DB_CREATE , 0) != 0) {
		printk(KERN_WARNING "i3fs: Unable to open access db");
	}
	mount_path[path_len] = '\0';
	
	strcat(mount_path, "access.db");
	if (accessdb->open(accessdb, NULL, mount_path, NULL, DB_BTREE, DB_CREATE , 0) != 0)     {
	printk(KERN_WARNING "i3fs: Unable to open access db");
	}

}


/*
 * Parse mount options: dir=XXX and debug=N
 *
 * Returns the dentry object of the lower-level (hidden) directory;
 * We want to mount our stackable file system on top of that hidden directory.
 *
 * Sets default debugging level to N, if any.
 */


dentry_t *
i3fs_parse_options(super_block_t *sb, char *options)
{
    dentry_t *hidden_root = ERR_PTR(-EINVAL);
    struct nameidata nd;
    char *name=NULL, *tmp, *end;
    int err = 0;
    void *checksum;
    char *dbdir=NULL;

    print_entry_location();
    
    stopd(sb)->allow_updates = 0;
    stopd(sb)->cache_results = 1;
    stopd(sb)->logfile = NULL;

    /* We don't want to go off the end of our arguments later on. */
    for (end = options; *end; end++);

    while (options < end) {
        tmp = options;
        while (*tmp && *tmp != ',')
            tmp++;
        *tmp = '\0';
        if (!strncmp("dir=", options, 4)) {
            /* note: the name passed need not be encoded */
            name = options + 4;
	    
	    stopd(sb)->mntdir = KMALLOC(strlen(name), GFP_KERNEL);
	    strcpy(stopd(sb)->mntdir, name);
	
            fist_dprint(4, "i3fs: using directory: %s\n", name);
            if (path_init(name, LOOKUP_FOLLOW, &nd))
                err = path_walk(name, &nd);
            if (err) {
                printk("i3fs: error accessing hidden directory '%s'\n", name);
                hidden_root = ERR_PTR(err);
                goto out;
            }
            hidden_root = nd.dentry;
            stopd(sb)->hidden_mnt = nd.mnt;
            fist_dprint(6, "parse_options: new s_root, inode: 0x%lx, 0x%lx\n",
                        (long) hidden_root, (long) hidden_root->d_inode);

	/* the passphrase for i3fs */
        } else if (!strncmp("pass=", options, 5)) {
	    stopd(sb)->passphrase = KMALLOC(strlen(options + 5), GFP_KERNEL);
            strcpy(stopd(sb)->passphrase, options + 5);
	    stopd(sb)->plen = strlen(stopd(sb)->passphrase);

        }
	/* custom path for storing the database files */
	else if (!strncmp("dbdir=", options, 6))	{
		dbdir = options + 6;
		if (path_init(dbdir, LOOKUP_FOLLOW, &nd))
	                err = path_walk(name, &nd);
		if (err)	{
			printk("i3fs: Error accessing db directory\n");
			dbdir = NULL;
		}
		printk("\nDB directory set to %s", dbdir);
	}
	/* custom log file where the i3fs report should be redirected */
	else if (!strncmp("logfile=", options, 8))	{
		name = options + 8;
		stopd(sb)->logfile = KMALLOC(strlen(options+8), GFP_KERNEL);
		strcpy(stopd(sb)->logfile, options+8);
		printk("\nLog file set to %s", name);
	}
	/* disable caching of results of no-cache option is set */
	else if	(!strncmp("no-cache=", options, 9))	{
		if (*(options + 9) == '1')	{ 
			stopd(sb)->cache_results = 0;
			printk("\ni3fs: Caching disabled");
		}
	}
	/* determine whether checksums have to be auto updated */
	else if (!strncmp("auto-update=", options, 12))      {
		if (*(options+12) == '1')	{
	                stopd(sb)->allow_updates = 1;
			printk("\ni3fs: Updates allowed");
		}
        }
	else if (!strncmp("debug=", options, 6)) {
            int debug = simple_strtoul(options + 6, NULL, 0);
            fist_set_debug_value(debug);
        } else {
            printk(KERN_WARNING "i3fs: unrecognized option '%s'\n", options);        	   }
        options = tmp + 1;
    }

	// Validate the passphrase now
    checksum = KMALLOC(CHECKSUM_SIZE, GFP_KERNEL);
    memset(checksum, 0, CHECKSUM_SIZE);

    compute_cksum("i3fspassphrase", 9, stopd(sb)->passphrase, stopd(sb)->plen, checksum);


    setup_databases(sb, dbdir);

    if (!verify_passphrase(stopd(sb)->policydb, checksum))	{
	hidden_root = (dentry_t*)(-5);
	goto out;
    }


out:

    print_exit_location();

    return hidden_root;
}




#ifdef FIST_MALLOC_DEBUG
/* for malloc debugging */
static atomic_t i3fs_malloc_counter;

void *
i3fs_KMALLOC(size_t len, int flag, int line, const char *file)
{
    void *ptr = (void *) KMALLOC(len, flag);
    if (ptr) {
	atomic_inc(&i3fs_malloc_counter);
	printk("KM:%d:%p:%d:%s\n", atomic_read(&i3fs_malloc_counter),ptr, line, file);
    }
    return ptr;
}

void
i3fs_KFREE(void *ptr, int line, const char *file)
{
    atomic_inc(&i3fs_malloc_counter);
    printk("KF:%d:%p:%d:%s\n", atomic_read(&i3fs_malloc_counter), ptr, line, file);
    KFREE(ptr);
}
#endif /* FIST_MALLOC_DEBUG */




/* for attach mode, we use a different ->read_super() in attach.c */
super_block_t *
i3fs_read_super(super_block_t *sb, void *raw_data, int silent)
{
    super_block_t *ret_sb = NULL;
    dentry_t *hidden_root;

    print_entry_location();

#ifdef FIST_MALLOC_DEBUG
    atomic_set(&i3fs_malloc_counter, 0); /* for malloc debugging */
#endif /* FIST_MALLOC_DEBUG */

    if (!raw_data) {
	printk(KERN_WARNING "i3fs_read_super: missing data argument\n");
	goto out;
    }
    /*
     * Allocate superblock private data
     */
    stopd(sb) = KMALLOC(sizeof(struct i3fs_sb_info), GFP_KERNEL);
    if (!stopd(sb)) {
	printk(KERN_WARNING "%s: out of memory\n", __FUNCTION__);
	goto out;
    }
    memset(stopd(sb), 0, sizeof(struct i3fs_sb_info));

    hidden_root = i3fs_parse_options(sb, raw_data);
    if (IS_ERR(hidden_root)) {
	printk(KERN_WARNING "i3fs_read_super: lookup_dentry failed (err = %ld)\n", PTR_ERR(hidden_root));
	goto out_free;
    }
    if (!hidden_root->d_inode) {
	printk(KERN_WARNING "i3fs_read_super: no directory to interpose on\n");
	goto out_dput;
    }
    stohs(sb) = hidden_root->d_sb;



    /*
     * Linux 2.4.2-ac3 and beyond has code in
     * mm/filemap.c:generic_file_write() that requires sb->s_maxbytes
     * to be populated.  If not set, all write()s under that sb will
     * return 0.
     *
     * Linux 2.4.4+ automatically sets s_maxbytes to MAX_NON_LFS;
     * the filesystem should override it only if it supports LFS.
     */
    /* non-SCA code is good to go with LFS */
    sb->s_maxbytes = hidden_root->d_sb->s_maxbytes;

    sb->s_op = &i3fs_sops;
    /*
     * we can't use d_alloc_root if we want to use
     * our own interpose function unchanged,
     * so we simply replicate *most* of the code in d_alloc_root here
     */
    sb->s_root = d_alloc(NULL, &(const struct qstr) { "/", 1, 0 });
    if (IS_ERR(sb->s_root)) {
	printk(KERN_WARNING "i3fs_read_super: d_alloc failed\n");
	goto out_dput;
    }

    sb->s_root->d_op = &i3fs_dops;
    sb->s_root->d_sb = sb;
    sb->s_root->d_parent = sb->s_root;
    /* link the upper and lower dentries */
    dtopd(sb->s_root) = (struct i3fs_dentry_info *) KMALLOC(sizeof(struct i3fs_dentry_info), GFP_KERNEL);
    if (!dtopd(sb->s_root)) {
	goto out_dput2;
    }
    dtohd(sb->s_root) = hidden_root;

    if (i3fs_interpose(hidden_root, sb->s_root, sb, 0))
	goto out_dput2;

	root_dentry = hidden_root;
	//if (root_dentry->d_inode) printk("Filename = %s\n", root_dentry->d_name.name);

    ret_sb = sb;
    fist_print_dentry("i3fs_read_super OUT hidden_dentry", hidden_root);
    fist_print_inode("i3fs_read_super OUT hidden_inode", hidden_root->d_inode);
	
/* Initialize and populate the I3FS databases */
	INIT_LIST_HEAD(&i3fs_db);
	INIT_LIST_HEAD(&negative_cache);

    // next line causes null ptr deref at mount(2) time
    // fist_print_dentry("%s OUT sb->s_root", __FUNCTION__, sb->s_root);
    goto out;

 out_dput2:
    dput(sb->s_root);
 out_dput:
    dput(hidden_root);
 out_free:
    if (stopd(sb)) {
	mntput(stopd(sb)->hidden_mnt);
    }
    KFREE(stopd(sb));
    stopd(sb) = NULL;
 out:
    fist_print_sb("OUT sb", sb);
    if (stopd(sb))
	fist_print_sb("OUT hidden_sb", stohs(sb));
    print_exit_location();


    return ret_sb;
}


/* wrapper to iunique to warn if we have a wrap-around */
ino_t
i3fs_iunique(struct super_block *sb, ino_t maxreserved)
{
    static ino_t counter = 0;
    ino_t ret;

    ret = iunique(sb, maxreserved);
    if (ret < counter) {
        printk(KERN_WARNING "%s iunique may have wrapped around!\n", __FUNCTION__);
    }
    counter = ret;

    return ret;
}


/*----*/
// this structure *must* be static!!! (or the memory for 'name' gets
// corrupted in 2.3...)
static DECLARE_FSTYPE(i3fs_fs_type, "i3fs", i3fs_read_super, 0);

static int __init init_i3fs_fs(void)
{
   printk("Registering i3fs version $Id: main.c,v 1.23 2004/08/19 23:40:48 gopalan Exp $\n");
	return register_filesystem(&i3fs_fs_type);;
}
static void __exit exit_i3fs_fs(void)
{
    printk("Unregistering i3fs version $Id: main.c,v 1.23 2004/08/19 23:40:48 gopalan Exp $\n");
    unregister_filesystem(&i3fs_fs_type);
}

EXPORT_NO_SYMBOLS;
MODULE_AUTHOR("Erez Zadok <ezk@cs.sunysb.edu>");
MODULE_DESCRIPTION("FiST-generated i3fs filesystem");
MODULE_LICENSE("GPL");

MODULE_PARM(fist_debug_var, "i");
MODULE_PARM_DESC(fist_debug_var, "Debug level");

module_init(init_i3fs_fs)
module_exit(exit_i3fs_fs)

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
